        .686p
	.xmm3
	.model	flat

	;;
	;; This set of definitions and macros support fixed time sorting
	;; of four or eight 32-bit floating point values and the 32-bit data
	;; items associated with each.
	;;
	;; The data structure assumed is an array of 32-bit values and
	;; 32-bit data identifiers as follows (in a C-style structure)
	;;
	;; struct sortdata {
	;;     FLOAT32 val;
	;;     UINT32  idx;  // idx is a 32 bit pointer or array offset
	;; } data[];
	;;

xmEQ	equ	0			; Equal, sorta useless
xmLT	equ	1			; Less than
xmLE	equ	2			; Less than or equal
xmUNORD	equ	3			; Unorderd (NAN involved), sorta useless
xmNEQ	equ	4			; Not equal, sorta useless
xmGE	equ	5			; Greater than or equal
xmGT	equ	6			; Greater than
xmORD	equ	7			; Ordered (neiter NAN), sorta useless

xmINT	equ	8			; Integer signed greater than

Temp1	equ	xmm6
Temp2	equ	xmm7

Order	equ	xmGT			; xmGT, xmINT for high to low sort


ordreg	macro	xreg1, xreg2, func, pflag
	movaps	Temp2, xreg1
	movaps	Temp1, xreg1
if (func - xmINT)			; Floating point compare?
	cmpps	Temp2, xreg2, func	; Yes, put mask in dest (TEMP2)
else
	pcmpgtd	Temp2, xreg2		; No, use integer GT and put mask
endif					; in dest (TEMP2)
	pxor	Temp1, xreg2		; XOR the register pair for swapping
ifnb <pflag>				; If indices carried in even registers
	pshufd	Temp2, Temp2, 10100000b	; copy XOR mask from data to pointer
endif
	pand	Temp1, Temp2		; If not swapping, zero mask

	pxor	xreg1, Temp1		; Now do the conditional swap
	pxor	xreg2, Temp1		; (on exit TEMP2 is still the swap mask
	endm				; for use inline by sort8 below)


prefix	macro	reg0, areg0

	;; The starting contents of the XMM registers are interleaved:

	;; i.e. ordered ptr1, data1, ptr0, data0,   =reg0
	;;              ptr3, data3, ptr2, data2,   =areg0

	pshufd	reg0, reg0, 11011000b
	pshufd	areg0, areg0, 11011000b

	;; now ordered ptr1, ptr0, data1, data0,
	;;             ptr3, ptr2, data3, data2,

	movaps	Temp1, areg0
	movhlps	areg0, reg0
	movlhps	reg0, Temp1

	;; Now structures are reordered for the sort algorithm
	;; i.e.:        data3, data2, data1, data0, =reg0
	;;              ptr3, ptr2, ptr1, ptr0      =areg0

	endm


sortpair macro	reg0, reg1, areg0, areg1, Order

	;; Sort the pairs vertically so each pair is in order

	ordreg	reg0, reg1, Order
	movaps	Temp1, areg0
	pxor	Temp1, areg1
	pand	Temp1, Temp2
	pxor	areg0, Temp1
	pxor	areg1, Temp1

	endm


lhxchg	macro	reg0, reg1, areg0, areg1
	;; Each data/pointer position is now sorted
	;; Now ordered max37, max26, max15, max04,
	;;             min37, min26, min15, min04

	movaps	Temp1, reg0
	movaps	Temp2, areg0
	movhlps	reg0, reg1
	movhlps	areg0, areg1
	movlhps	reg1, Temp1
	movlhps	areg1, Temp2

	;; Now ordered max37, max26, min37, min26,
	;;             max15, max04, min15, min04
	endm


grab8	macro	reg0, reg1, areg0, areg1

	;; now ordered data3, data2, data1, data0,
	;;             ptr3, ptr2, ptr1, ptr0
	;;             data7, data6, data5, data4,
	;;             ptr7, ptr6, ptr5, ptr4

	sortpair reg0, reg1, areg0, areg1, Order

	lhxchg	reg0, reg1, areg0, areg1
	sortpair reg0, reg1, areg0, areg1, Order

	;; Now ordered max1357, max0246, mid1357, mid0246,
	;;             mid1357, mid0246, min1357, min0246

	pshufd	reg1, reg1, 01001110b
	pshufd	areg1, areg1, 01001110b

	;; Now ordered max1357, max0246, mid1357, mid0246,
	;;             min1357, min0246, mid1357, mid0246

	sortpair reg0, reg1, areg0, areg1, Order

	;; Now ordered max1357, max0246, hih1357, hih0246,
	;;             min1357, min0246, low1357, low0246

	pshufd	reg0, reg0, 11011000b
	pshufd	areg0, areg0, 11011000b

	pshufd	reg1, reg1, 01110010b
	pshufd	areg1, areg1, 01110010b

	;; Now ordered max1357,	hih1357, max0246, hih0246,
	;;             low1357, min1357, low0246, min0246

	lhxchg	reg0, reg1, areg0, areg1
	sortpair reg0, reg1, areg0, areg1, Order

	;; Now ordered sort7,	hih1357, low1357, min1357,
	;;             max0246, hih0246, low0246, sort0

	pshufd	reg0, reg0, 10010011b
	pshufd	areg0, areg0, 10010011b

	;; Now ordered hih1357, low1357, min1357, sort7
	;;             max0246, hih0246, low0246, sort0

	sortpair reg0, reg1, areg0, areg1, Order

	;; Now ordered sort6,   low1357, min1357, sort7
	;;             max0246, hih0246, sort1,   sort0

	pshufd	reg0, reg0, 10010011b
	pshufd	areg0, areg0, 10010011b

	;; Now ordered low1357, min1357, sort7, sort6
	;;             max0246, hih0246, sort1, sort0

	sortpair reg0, reg1, areg0, areg1, Order

	;; Now ordered sort5,   min1357, sort7, sort6
	;;             max0246, sort2,   sort1, sort0

	pshufd	reg0, reg0, 10010011b
	pshufd	areg0, areg0, 10010011b

	;; Now ordered min1357, sort7, sort6, sort5
	;;             max0246, sort2, sort1, sort0

	sortpair reg0, reg1, areg0, areg1, Order

	;; Now ordered sort4, sort7, sort6, sort5
	;;             sort3, sort2, sort1, sort0

	pshufd	reg0, reg0, 10010011b
	pshufd	areg0, areg0, 10010011b

	;; Now ordered sort7, sort6, sort5, sort4
	;;             sort3, sort2, sort1, sort0
	;; and finally write out the results!

	endm


	.code

	public	_sort4
_sort4	proc
	push	ebx
	mov	ebx, ss:[esp+8]

	movups	xmm0, ds:[ebx]
	movups	xmm1, ds:[ebx+16]

	;; now ordered ptr1, data1, ptr0, data0,
	;;             ptr3, data3, ptr2, data2,

	ordreg	xmm0, xmm1, Order, 1

	;; now ordered ptr13a, max13, ptr02a, max02,
	;;             ptr13b, min13, ptr02b, min02,
	
	movaps	Temp2, xmm0		; exchange low 2 dw of xmm0 with
	movlhps	xmm0, xmm1		; the high 2 dw of xmm1
	movhlps	xmm1, Temp2

	;; now ordered ptr13a, max13, ptr13b, min13,
	;;             ptr02a, max02, ptr02b, min02,

	ordreg	xmm0, xmm1, Order, 1

	;; now ordered *, sort3, *, midb,
	;;             *, mida,  *, sort0,

	pshufd	xmm1, xmm1, 01001110b

	;; now ordered *, sort3, *, midb,
	;;             *, sort0, *, mida,

	ordreg	xmm0, xmm1, Order, 1

	;; now ordered *, sort3, *, sort2,
	;;             *, sort0, *, sort1,

	pshufd	xmm0, xmm0, 01001110b

	;; now ordered *, sort2, *, sort3,
	;;             *, sort0, *, sort1,

	movups	ds:[ebx+16], xmm0
	movups	ds:[ebx], xmm1

	pop	ebx
	ret
_sort4	endp


	public	_sort8
_sort8	proc
	push	ebx
	mov	ebx, ss:[esp+8]

	;; First, load the data into XMM registers

	movups	xmm0, ds:[ebx]
	movups	xmm2, ds:[ebx+16]	
	movups	xmm1, ds:[ebx+32]
	movups	xmm3, ds:[ebx+48]

	;; Second, sort the 8 data elements

	prefix	xmm0, xmm1
	prefix	xmm2, xmm3
	grab8	xmm0, xmm2, xmm1, xmm3

	;; Third, re-interleave the sort keys and data pointers
	;; and store the resultant structures back into memory

	movaps	Temp2, xmm2
	unpcklps Temp2, xmm3
	movups	ds:[ebx], Temp2

	movaps	Temp2, xmm2
	unpckhps Temp2, xmm3
	movups	ds:[ebx+16], Temp2

	movaps	Temp2, xmm0
	unpcklps Temp2, xmm1
	movups	ds:[ebx+32], Temp2

	movaps	Temp2, xmm0
	unpckhps Temp2, xmm1
	movups	ds:[ebx+48], Temp2

	;; Restore EBX and exit the sort routine

	pop	ebx
	ret
_sort8	endp


	public	_sort12
_sort12	proc
	push	ebx
	mov	ebx, ss:[esp+8]

	;; First, load the data into XMM registers

	movups	xmm0, ds:[ebx]
	movups	xmm3, ds:[ebx+16]	
	movups	xmm1, ds:[ebx+32]
	movups	xmm4, ds:[ebx+48]
	movups	xmm2, ds:[ebx+64]
	movups	xmm5, ds:[ebx+80]

	;; Second, reorder the 12 data elements

	prefix	xmm0, xmm3
	prefix	xmm1, xmm4
	prefix	xmm2, xmm5

	;; Next bisect the 8 values in XMM0 and XMM2
	;;   we could use
	;;   == grab8  xmm2, xmm0, xmm5, xmm3
	;;   but we just need put higher half on one side
	;;   and the lower half on the other side

	sortpair xmm2, xmm0, xmm5, xmm3, Order
	pshufd	xmm0, xmm0, 10010011b
	pshufd	xmm3, xmm3, 10010011b
	sortpair xmm2, xmm0, xmm5, xmm3, Order
	pshufd	xmm0, xmm0, 10010011b
	pshufd	xmm3, xmm3, 10010011b
	sortpair xmm2, xmm0, xmm5, xmm3, Order
	pshufd	xmm0, xmm0, 10010011b
	pshufd	xmm3, xmm3, 10010011b
	sortpair xmm2, xmm0, xmm5, xmm3, Order

	;; Third, sort the values in each group of 4
	;; with the remaining 4 unbinned entries

	grab8	xmm2, xmm1, xmm5, xmm4
	grab8	xmm1, xmm0, xmm4, xmm3

	;; Finally return the sorted entries to the buffer

	movaps	Temp1, xmm0
	movaps	Temp2, xmm0
	unpcklps Temp1, xmm3
	unpckhps Temp2, xmm3
	movups	ds:[ebx], Temp1
	movups	ds:[ebx+16], Temp2

	movaps	Temp1, xmm1
	movaps	Temp2, xmm1
	unpcklps Temp1, xmm4
	unpckhps Temp2, xmm4
	movups	ds:[ebx+32], Temp1
	movups	ds:[ebx+48], Temp2

	movaps	Temp1, xmm2
	movaps	Temp2, xmm2
	unpcklps Temp1, xmm5
	unpckhps Temp2, xmm5
	movups	ds:[ebx+64], Temp1
	movups	ds:[ebx+80], Temp2

	;; Restore EBX and exit the sort routine

	pop	ebx
	ret
_sort12	endp

	end
